/*
	matched.c
	
	Copyright (C) 2003 Tim Goetze <tim@quitte.de>
	
	http://quitte.de/dsp/

	16th order IIR filter modeled after an impulse response from
	a matchless chieftain guitar amplifier, corrected for the
	use in conjunction with the 'preamp' (see url above) plugin.
*/
/*
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
	02111-1307, USA or point your web browser to http://www.gnu.org.
*/

#include <malloc.h>
#include <ladspa.h>

typedef struct 
{
	LADSPA_Data * in;
	LADSPA_Data * out;

	LADSPA_Data gain;

	int z;
	LADSPA_Data x[16], y[16];
} 
matched;

#define PROCESS() \
	a *= +1.09170598489534098618; \
	a += +0.09079972214735353153 * x [(z-1) & 15]; \
	a += +0.25594415074966925250 * y [(z-1) & 15]; \
	a += +0.06275851860674004634 * x [(z-2) & 15]; \
	a += +0.20269069460094527035 * y [(z-2) & 15]; \
	a += +0.01060723399364926073 * x [(z-3) & 15]; \
	a += +0.09969966728889961827 * y [(z-3) & 15]; \
	a += -0.04003298804448884862 * x [(z-4) & 15]; \
	a += -0.01698308894227101157 * y [(z-4) & 15]; \
	a += -0.05098733387148121632 * x [(z-5) & 15]; \
	a += -0.10025666121442594614 * y [(z-5) & 15]; \
	a += -0.00212027924805496424 * x [(z-6) & 15]; \
	a += -0.12376947415521612206 * y [(z-6) & 15]; \
	a += +0.08216861428657665223 * x [(z-7) & 15]; \
	a += -0.09837065464554287098 * y [(z-7) & 15]; \
	a += +0.14558022807771681117 * x [(z-8) & 15]; \
	a += -0.05792397408788292623 * y [(z-8) & 15]; \
	a += +0.14107979919296464999 * x [(z-9) & 15]; \
	a += -0.02742170520254189059 * y [(z-9) & 15]; \
	a += +0.06863603090258478534 * x [(z-10) & 15]; \
	a += -0.00754151013889273192 * y [(z-10) & 15]; \
	a += -0.02221952179046714215 * x [(z-11) & 15]; \
	a += +0.01411725095447000089 * y [(z-11) & 15]; \
	a += -0.06830834967903634536 * x [(z-12) & 15]; \
	a += +0.03438707838281616913 * y [(z-12) & 15]; \
	a += -0.04122301442705932767 * x [(z-13) & 15]; \
	a += +0.02016038635935246681 * y [(z-13) & 15]; \
	a += +0.03204378656092733701 * x [(z-14) & 15]; \
	a += -0.07309957271207177809 * y [(z-14) & 15]; \
	a += +0.09023862342328989317 * x [(z-15) & 15]; \
	a += -0.25732881619369524229 * y [(z-15) & 15]; 

/* /////////////////////////////////////////////////////////////////////// */

static LADSPA_Handle
instantiate (const struct _LADSPA_Descriptor * desc, unsigned long sr)
{
	matched * h = (matched *) malloc (sizeof (matched));
	h->gain = 1.;
	return h;
}

static void
connect_port (LADSPA_Handle h, unsigned long n, LADSPA_Data * to)
{
	if (n == 0)
		((matched *) h)->in = to;
	else if (n == 1)
		((matched *) h)->out = to;
}

static void
activate (LADSPA_Handle h)
{
	LADSPA_Data * x = ((matched *) h)->x;
	LADSPA_Data * y = ((matched *) h)->y;

	int i = 0;

	while (i < 16)
		x[i] = 
		y[i] = 0.,
		++i;

	((matched *) h)->z = 0;
}

static void
run (LADSPA_Handle h, unsigned long frames)
{
	unsigned long i;

	register LADSPA_Data * x = ((matched *) h)->x;
	register LADSPA_Data * y = ((matched *) h)->y;

	register LADSPA_Data * in = ((matched *) h)->in;
	register LADSPA_Data * out = ((matched *) h)->out;

	register int z = ((matched *) h)->z;

	for (i = 0; i < frames; ++i)
	{
		register LADSPA_Data a = in[i];
		
		x[z] = a;
		
		PROCESS();

		y[z] = a;
		
		out[i] = (LADSPA_Data) a;

		z = (z + 1) & 15;
	}

	((matched *) h)->z = z;
}

static void
run_adding (LADSPA_Handle h, unsigned long frames)
{
	unsigned long i;

	register LADSPA_Data * x = ((matched *) h)->x;
	register LADSPA_Data * y = ((matched *) h)->y;

	register LADSPA_Data * in = ((matched *) h)->in;
	register LADSPA_Data * out = ((matched *) h)->out;

	register LADSPA_Data gain = ((matched *) h)->gain;

	register int z = ((matched *) h)->z;

	for (i = 0; i < frames; ++i)
	{
		register LADSPA_Data a = in[i];
		
		x[z] = a;
		
		PROCESS();

		y[z] = a;
		
		out[i] += (LADSPA_Data) (gain * a);

		z = (z + 1) & 15;
	}

	((matched *) h)->z = z;
}

static void
set_run_adding_gain (LADSPA_Handle h, LADSPA_Data gain)
{
	((matched *) h)->gain = gain;
}

static void
cleanup (LADSPA_Handle h)
{
	free (h);
}

/* /////////////////////////////////////////////////////////////////////// */

static const LADSPA_PortDescriptor 
port_descriptors[] = 
{
	LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO,
	LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO,
};

static const char *
port_names[] =
{
	"in",
	"out",
};

static const LADSPA_PortRangeHint
port_range_hints[] =
{
	/* gaan we mieren neuken? ja wel! */
	{
		LADSPA_HINT_BOUNDED_BELOW | 
		LADSPA_HINT_BOUNDED_ABOVE | 
		LADSPA_HINT_DEFAULT_MIDDLE,
		-1.0, 1.0
	},
	{
		LADSPA_HINT_BOUNDED_BELOW | 
		LADSPA_HINT_BOUNDED_ABOVE | 
		LADSPA_HINT_DEFAULT_MIDDLE,
		-1.0, 1.0
	},
};

static const LADSPA_Descriptor
descriptor =
{
	UniqueID:	1766,
	Label: "matched",
	Properties: LADSPA_PROPERTY_HARD_RT_CAPABLE,
	Name: "matched (amp tone)",
	Maker: "Tim Goetze <tim@quitte.de>",
	Copyright: "GPL, 2003 Tim Goetze",
	
	PortCount: 2,
	PortDescriptors: port_descriptors,
	PortNames: port_names,
	PortRangeHints: port_range_hints,
	
	/* methods */
	instantiate: instantiate,
	connect_port: connect_port,
	activate: activate,
	run: run,
	run_adding: run_adding,
	set_run_adding_gain: set_run_adding_gain,
	cleanup: cleanup,
};

/* /////////////////////////////////////////////////////////////////////// */

const LADSPA_Descriptor *
ladspa_descriptor (unsigned long index) 
{
	return index ? 0 : &descriptor;
}

